#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Any, Callable, Dict, Generator, Tuple, Union, Optional

ProbeGenType = Generator[float, None, Optional[bool]]
ProbeGenFnType = Callable[..., ProbeGenType]
ProbeRegularFnType = Callable[..., Any]
ProbeFnType = Union[ProbeGenFnType, ProbeRegularFnType]

ForgeGenType = Generator[Any, None, None]
ForgeGenFnType = Callable[..., ForgeGenType]
ForgeRegularFnType = Callable[..., Any]
ForgeFnType = Union[ForgeGenFnType, ForgeRegularFnType]

TestFnType = Callable[..., Any]

ArtifactsType = Dict[str, Any]
ExecutableKeyType = Tuple[str, ...]
