#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging

UCC_MODINPUT_TEST_LOGGER_NAME = "splunk-add-on-ucc-modinput-test-functional"
UCC_MODINPUT_TEST_LOG_FILE_NAME = f"{UCC_MODINPUT_TEST_LOGGER_NAME}.log"

logFormatter = logging.Formatter(
    "%(asctime)s,%(msecs)03d %(levelname)s %(name)s pid=%(process)d tid=%(thread)d file=%(filename)s func=%(funcName)s line=%(lineno)d %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

fileHandler = logging.FileHandler(UCC_MODINPUT_TEST_LOG_FILE_NAME)
fileHandler.setFormatter(logFormatter)

logger = logging.getLogger(UCC_MODINPUT_TEST_LOGGER_NAME)
logger.addHandler(fileHandler)
logger.setLevel(logging.DEBUG)
